@@ -105,7 +105,7 @@ const ReplaySession::MemoryRanges& ReplaySession::always_free_address_space(
     auto frame = tmp_reader.read_frame();
     auto event = frame.event();
     // If a region was ever mprotected to something that's not PROT_NONE,
-    // we need to delete it as well.
+    // or had PR_SET_VMA_ANON_NAME called on it, we need to delete it as well.
     if (event.is_syscall_event()) {
       auto syscall_event = event.Syscall();
       if (is_mprotect_syscall(syscall_event.number, syscall_event.arch()) ||
@@ -117,6 +117,14 @@ const ReplaySession::MemoryRanges& ReplaySession::always_free_address_space(
           delete_range(*result, MemoryRange(start, size));
         }
       }
+      if (is_prctl_syscall(syscall_event.number, syscall_event.arch())) {
+        auto regs = frame.regs();
+        if (regs.arg2() == PR_SET_VMA_ANON_NAME) {
+          remote_ptr<void> start = regs.arg3();
+          size_t size = regs.arg4();
+          delete_range(*result, MemoryRange(start, size));
+        }
+      }
     } else if (event.is_syscallbuf_flush_event()) {
       auto syscallbuf_flush_event = event.SyscallbufFlush();
       for (auto& record : syscallbuf_flush_event.mprotect_records) {
@@ -262,6 +270,8 @@ ReplaySession::ReplaySession(const std::string& dir, const Flags& flags)
   }
 
   set_intel_pt_enabled(flags.intel_pt_start_checking_event >= 0);
+
+  check_virtual_address_size();
 }
 
 ReplaySession::ReplaySession(const ReplaySession& other)
@@ -295,6 +305,22 @@ ReplaySession::~ReplaySession() {
   DEBUG_ASSERT(emufs().size() == 0);
 }
 
+void ReplaySession::check_virtual_address_size() const
+{
+  uint8_t virtual_address_size_needed = trace_in.max_virtual_address_size();
+  if (virtual_address_size_supported(virtual_address_size_needed)) {
+    return;
+  }
+
+  if (rr::Flags::get().force_things) {
+    LOG(warn) << "Virtual address size is unsupported but forcing anyways.";
+    return;
+  }
+
+  CLEAN_FATAL() << "The trace uses " << (uint32_t)virtual_address_size_needed <<
+      " bit virtual addresses but this system does not support that size.";
+}
+
 ReplaySession::shr_ptr ReplaySession::clone() {
   LOG(debug) << "Deepforking ReplaySession " << this << " ...";
 
@@ -489,7 +515,7 @@ bool ReplaySession::handle_unrecorded_cpuid_fault(
     ReplayTask* t, const StepConstraints& constraints) {
   if (t->stop_sig() != SIGSEGV || !has_cpuid_faulting() ||
       trace_in.uses_cpuid_faulting() ||
-      trapped_instruction_at(t, t->ip()) != TrappedInstruction::CPUID) {
+      special_instruction_at(t, t->ip()).opcode != SpecialInstOpcode::X86_CPUID) {
     return false;
   }
   // OK, this is a case where we did not record using CPUID faulting but we are
@@ -514,7 +540,7 @@ bool ReplaySession::handle_unrecorded_cpuid_fault(
               << " CX=" << HEX(r.cx()) << "; defaulting to 0/0/0/0";
     r.set_cpuid_output(0, 0, 0, 0);
   }
-  r.set_ip(r.ip() + trapped_instruction_len(TrappedInstruction::CPUID));
+  r.set_ip(r.ip() + special_instruction_len(SpecialInstOpcode::X86_CPUID));
   t->set_regs(r);
   // Clear SIGSEGV status since we're handling it
   t->set_status(constraints.is_singlestep() ? WaitStatus::for_stop_sig(SIGTRAP)
@@ -982,6 +1008,7 @@ Completion ReplaySession::emulate_async_signal(
    * be dealt with. */
   bool pending_SIGTRAP = false;
   bool did_set_internal_breakpoints = false;
+  bool did_set_bpf_breakpoint = false;
   RunCommand SIGTRAP_run_command = RUN_CONTINUE;
 
   /* Step 2: more slowly, find our way to the target ticks and
@@ -1042,17 +1069,21 @@ Completion ReplaySession::emulate_async_signal(
         // breakpoint instruction in the tracee would have triggered a
         // deterministic signal instead of an async one.
         // So we must have hit our internal breakpoint.
-        ASSERT(t, did_set_internal_breakpoints);
+        ASSERT(t, did_set_internal_breakpoints || did_set_bpf_breakpoint);
         // We didn't do an internal singlestep, and if we'd done a
         // user-requested singlestep we would have hit the above case.
         ASSERT(t, !trap_reasons.singlestep);
-        if (t->ip().undo_executed_bkpt(t->arch()) == in_syscallbuf_syscall_hook) {
-          t->vm()->remove_breakpoint(ip, BKPT_INTERNAL);
-          t->vm()->remove_breakpoint(in_syscallbuf_syscall_hook, BKPT_INTERNAL);
-          t->move_ip_before_breakpoint();
-          return COMPLETE;
+        if (did_set_internal_breakpoints) {
+          if (t->ip().undo_executed_bkpt(t->arch()) == in_syscallbuf_syscall_hook) {
+            t->vm()->remove_breakpoint(ip, BKPT_INTERNAL);
+            t->vm()->remove_breakpoint(in_syscallbuf_syscall_hook, BKPT_INTERNAL);
+            t->move_ip_before_breakpoint();
+            return COMPLETE;
+          }
+          ASSERT(t, regs.ip() == t->ip().undo_executed_bkpt(t->arch()));
+        } else {
+          LOG(debug) << "    fast-forwarded through " << t->hpc.bpf_skips() << " breakpoint hits with bpf";
         }
-        ASSERT(t, regs.ip() == t->ip().undo_executed_bkpt(t->arch()));
         /* Case (1) above: cover the tracks of
          * our internal breakpoint, and go
          * check again if we're at the
@@ -1060,7 +1091,9 @@ Completion ReplaySession::emulate_async_signal(
         LOG(debug) << "    trap was for target $ip";
 
         pending_SIGTRAP = false;
-        t->move_ip_before_breakpoint();
+        if (did_set_internal_breakpoints) {
+          t->move_ip_before_breakpoint();
+        }
         /* We just backed up the $ip, but
          * rewound it over an |int $3|
          * instruction, which couldn't have
@@ -1093,6 +1126,7 @@ Completion ReplaySession::emulate_async_signal(
       }
       did_set_internal_breakpoints = false;
     }
+    did_set_bpf_breakpoint = false;
 
     if (at_target) {
       /* Case (2) above: done. */
@@ -1117,11 +1151,16 @@ Completion ReplaySession::emulate_async_signal(
        * no slower than single-stepping our way to
        * the target execution point. */
       LOG(debug) << "    breaking on target $ip";
-      t->vm()->add_breakpoint(ip, BKPT_INTERNAL);
-      if (in_syscallbuf_syscall_hook) {
-        t->vm()->add_breakpoint(in_syscallbuf_syscall_hook, BKPT_INTERNAL);
+      if (is_x86_string_instruction_at(t, ip) || !t->hpc.accelerate_async_signal(regs)) {
+        t->vm()->add_breakpoint(ip, BKPT_INTERNAL);
+
+        if (in_syscallbuf_syscall_hook) {
+          t->vm()->add_breakpoint(in_syscallbuf_syscall_hook, BKPT_INTERNAL);
+        }
+        did_set_internal_breakpoints = true;
+      } else {
+        did_set_bpf_breakpoint = true;
       }
-      did_set_internal_breakpoints = true;
       continue_or_step(t, constraints, RESUME_UNLIMITED_TICKS);
       SIGTRAP_run_command = constraints.command;
     } else {
@@ -1807,6 +1846,11 @@ ReplayTask* ReplaySession::setup_replay_one_trace_frame(ReplayTask* t) {
     t = revive_task_for_exec();
   }
 
+  if (t->tuid() != last_task_tuid) {
+    t->will_schedule();
+    last_task_tuid = t->tuid();
+  }
+
   LOG(debug) << "[event " << trace_frame.time() << "] " << t->rec_tid
              << ": replaying " << Event(ev) << "; state "
              << (ev.is_syscall_event() ? state_name(ev.Syscall().state)
@@ -2196,9 +2240,14 @@ void ReplaySession::reattach_tasks(ScopedFd new_tracee_socket, ScopedFd new_trac
     if (!t->wait()) {
       FATAL() << "Task " << t->tid << " killed unexpectedly";
     }
-    if (SIGSTOP != t->status().group_stop()) {
-      WaitStatus failed_status = t->status();
-      FATAL() << "Unexpected stop " << failed_status << " for " << t->tid;
+    WaitStatus status = t->status();
+    // Normally the SIGSTOP from detach_tasks() will have been delivered to the tracee
+    // while it was detached, putting it into a group stop, so we'll see the group stop
+    // status here. However it is possible for the SIGSTOP to be queued but not delivered
+    // because the tracee hasn't been scheduled yet. In that case we might see the
+    // SIGSTOP signal stop here instead.
+    if (status.group_stop() != SIGSTOP && status.stop_sig() != SIGSTOP) {
+      FATAL() << "Unexpected stop " << status << " for " << t->tid;
     }
     t->clear_wait_status();
     t->open_mem_fd();
@@ -2211,7 +2260,7 @@ bool ReplaySession::mark_stdio() const {
 }
 
 bool ReplaySession::echo_stdio() const {
-  return flags().redirect_stdio &&
+  return flags().redirect_stdio && visible_execution_ &&
     current_frame_time() >= suppress_stdio_before_event_;
 }
 
